#include "FFVideoEncoder.h"
#include "fillplugins.h"
#include <string.h>
#include <stdio.h>

AVM_BEGIN_NAMESPACE;

FFVideoEncoder::FFVideoEncoder(AVCodec* av, const CodecInfo& info, fourcc_t compressor, const BITMAPINFOHEADER& bh)
    :IVideoEncoder(info), m_pAvCodec(av), m_pAvContext(0), m_Caps(CAP_YV12),
    m_bh(bh), m_obh(bh)
{
    m_obh.biCompression = info.fourcc;
    m_obh.biHeight = labs(m_obh.biHeight);
}

FFVideoEncoder::~FFVideoEncoder()
{
    Stop();
}

int FFVideoEncoder::EncodeFrame(const CImage* src, void* dest, int* is_keyframe, uint_t* size, int* lpckid)
{
    int hr = 0;

    if (!m_pAvContext)
    {
	m_pAvContext = avcodec_alloc_context();
        m_pAvContext->width = m_bh.biWidth;
	m_pAvContext->height = m_obh.biHeight;
	//m_pAvContext->pix_fmt = PIX_FMT_YUV420P;

	m_pAvContext->bit_rate = 1000000;
	m_pAvContext->frame_rate = 25;
	m_pAvContext->frame_rate_base = 1;
	m_pAvContext->gop_size = 250;
	m_pAvContext->qmin = 2;
	m_pAvContext->qmax = 31;
	//m_pAvContext->flags = CODEC_FLAG_HQ;

        printf("CODEC opening  %dx%d\n", m_bh.biWidth, m_obh.biHeight);

	if (avcodec_open(m_pAvContext, m_pAvCodec) < 0)
	{
	    free(m_pAvContext);
	    m_pAvContext = 0;
	    return -1;
	}
#if 0
	switch (t) {
	case FF_OPT_TYPE_BOOL:
	    i += sprintf(def + i, "%s%s=%s",
			 col, c->name,
			 c->defval != 0. ? "on" : "off");
	    break;
	case FF_OPT_TYPE_DOUBLE:
	    i += sprintf(def + i, "%s%s=%f",
			 col, c->name, c->defval);
	    break;
	case FF_OPT_TYPE_INT:
	    i += sprintf(def + i, "%s%s=%d",
			 col, c->name, (int) c->defval);
	    break;
	case FF_OPT_TYPE_STRING:
	    if (c->defstr) {
		char* d = av_strdup(c->defstr);
		char* f = strchr(d, ',');
		if (f)
		    *f = 0;
		i += sprintf(def + i, "%s%s=%s",
			     col, c->name, d);
		av_free(d);
	    }
	    break;
	}
	col = ":";
	c++;
#endif


    }

    CImage* ci = 0;
    //printf("CODEC format  0x%x\n", src->Format());
    if (src->Format() != IMG_FMT_YV12)
    {
        printf("Converted\n");
        ci = new CImage(src, IMG_FMT_YV12);
    }
    else
        ci = (CImage*) src;

    AVFrame f; // YV12 only
    memset(&f, 0, sizeof(AVFrame));
    f.data[0] = ci->Data(0);
    f.data[1] = ci->Data(2);
    f.data[2] = ci->Data(1);
    f.linesize[0] = src->Stride(0);
    f.linesize[1] = src->Stride(2);
    f.linesize[2] = src->Stride(1);
    //printf("ECDING FF  %p %p %p   sz:%d\n", f.data[0], f.data[1], f.data[2], GetOutputSize());
    //printf("ECDING FF  %d %d %d\n", f.linesize[0], f.linesize[1], f.linesize[2]);

    int rsize = avcodec_encode_video(m_pAvContext, (unsigned char*) dest,
				     GetOutputSize(), &f);
    //printf("ECDING FF  size %d\n", rsize);
    if (size)
	*size = rsize;
    if (is_keyframe) {
	*is_keyframe = m_pAvContext->coded_frame->key_frame;
	//printf("KEYFRAME %d\n", *is_keyframe);
    }
    if (ci != src)
        ci->Release();

    return hr;
}

int FFVideoEncoder::Start()
{
    return 0;
}

int FFVideoEncoder::Stop()
{
    if (m_pAvContext)
    {
	avcodec_close(m_pAvContext);
	free(m_pAvContext);
        m_pAvContext = 0;
    }

    return 0;
}

int FFVideoEncoder::SetQuality(int quality)
{
    if (quality < 0 || quality >10000)
	return - 1;
    m_iQual = quality;

    return 0;
}

int FFVideoEncoder::SetKeyFrame(int freq)
{
    return 0;
}

AVM_END_NAMESPACE;
